"""
Contains the CategorizedKeyExtract and a schema for (de)serialization.
"""

from itertools import combinations, product
from typing import List

import attrs
from marshmallow import Schema, fields, post_load

from ahbicht.models.condition_nodes import ConditionFulfilledValue, EvaluatedFormatConstraint
from ahbicht.models.content_evaluation_result import ContentEvaluationResult


# pylint: disable=too-few-public-methods,  unused-argument
@attrs.define(auto_attribs=True)
class CategorizedKeyExtract:
    """
    A Categorized Key Extract contains those condition keys that are contained inside an expression.
    For expressions (that do not contain any unresolved package) it's possible to pre-generate all possible outcomes of
    a content evaluation. CategorizedKeyExtract is also the answer to the question:
    'Which information do I need to provide in a ContentEvaluationResult in order to evaluate a given expression?'
    """

    #: list of keys for which you'll need to provide hint texts in a ContentEvaluationResult
    hint_keys: List[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(str), iterable_validator=attrs.validators.instance_of(list)
        )
    )
    #: list of keys for which you'll need to provide EvaluatedFormatConstraints
    format_constraint_keys: List[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(str), iterable_validator=attrs.validators.instance_of(list)
        )
    )

    #: list of keys for which you'll need to provide ConditionFulfilledValues
    requirement_constraint_keys: List[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(str), iterable_validator=attrs.validators.instance_of(list)
        )
    )

    #: list of packages that need to be resolved (additionally)
    package_keys: List[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.matches_re(r"^\d+P$"),
            iterable_validator=attrs.validators.instance_of(list),
            # todo: implement wiederholbarkeiten
        )
    )

    #: a list of time conditions, if present
    time_condition_keys: List[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.matches_re(r"^UB(?:1|2|3)$"),
            iterable_validator=attrs.validators.instance_of(list),
        )
    )

    def _remove_duplicates(self) -> None:
        """
        remove duplicates from all lists
        """
        self.hint_keys = [*set(self.hint_keys)]
        self.format_constraint_keys = [*set(self.format_constraint_keys)]
        self.requirement_constraint_keys = [*set(self.requirement_constraint_keys)]
        self.package_keys = [*set(self.package_keys)]
        self.time_condition_keys = [*set(self.time_condition_keys)]

    def _sort_keys(self) -> None:
        """
        sort the keys in all lists in ascending order
        """
        self.hint_keys.sort(key=int)
        self.format_constraint_keys.sort(key=int)
        self.requirement_constraint_keys.sort(key=self._sort_repeat_key)
        self.package_keys.sort()
        self.time_condition_keys.sort()

    def _sort_repeat_key(self, item: str) -> tuple:
        """
        Custom sort key for requirement_constraint_keys to handle 'n..m' format.
        """
        if ".." in item:
            n, m = map(int, item.split(".."))
            return (n, m)
        return (int(item),)

    def __add__(self, other: "CategorizedKeyExtract") -> "CategorizedKeyExtract":
        """
        Joins both this and the other categorized key extract.
        The result returned contains all those keys (once) that are present in either one or both of the summands.
        Duplicates will be removed.
        :param other: another categorized key extract
        :return: a new categorized key extract with the keys from both summands (which will be left untouched)
        """
        result = CategorizedKeyExtract(
            hint_keys=self.hint_keys + other.hint_keys,
            requirement_constraint_keys=self.requirement_constraint_keys + other.requirement_constraint_keys,
            format_constraint_keys=self.format_constraint_keys + other.format_constraint_keys,
            package_keys=self.package_keys + other.package_keys,
            time_condition_keys=self.time_condition_keys + other.time_condition_keys,
        )
        result.sanitize()  # removes duplicates, sorts the keys
        return result

    def sanitize(self) -> None:
        """
        Sanitize the result (remove duplicates, sort keys)
        """
        self._remove_duplicates()
        self._sort_keys()

    def generate_possible_content_evaluation_results(self) -> List[ContentEvaluationResult]:
        """
        A categorized key extract allows generating nearly all possible content evaluation results,
        except for hints, error messages, resolving packages.
        """
        results: List[ContentEvaluationResult] = []
        if len(self.format_constraint_keys) == 0 and len(self.requirement_constraint_keys) == 0:
            return results
        # for easier debugging below, replace the generators "(" with a materialized lists "["
        if len(self.format_constraint_keys) > 0:
            possible_fcs = (
                z
                for z in combinations(
                    product(self.format_constraint_keys, [True, False]), len(self.format_constraint_keys)
                )
                # y[0] is the key (and y[1] is the value (true/false))
                # We only want those combinations where the number of distinct keys == the number of keys present.
                # To better understand this, try evaluating the following expression in your debugger:
                #   list(combinations(product(["A", "B", "C"], [True, False]), len(["A", "B", "C"])))
                # In this (unfiltered) result, you'll find entries with the keys: ('A', 'A', 'B') or ('A', 'B', 'B').
                # These artefacts will be removed with the following if. Only the ('A', 'B', 'C') entries will pass.
                if len({y[0] for y in z}) == len(self.format_constraint_keys)
            )
        else:
            possible_fcs = [(("fc_dummy", True),)]  # type:ignore[assignment]

        if len(self.requirement_constraint_keys) > 0:
            possible_rcs = (
                z
                for z in combinations(
                    product(self.requirement_constraint_keys, ConditionFulfilledValue),
                    len(self.requirement_constraint_keys),
                )
                if len({y[0] for y in z}) == len(self.requirement_constraint_keys)
            )
        else:
            possible_rcs = [(("rc_dummy", ConditionFulfilledValue.NEUTRAL),)]  # type:ignore[assignment]

        for fc_rc_tuple in product(possible_fcs, possible_rcs):
            # This product would have length 0 if one of the "factors" had length 0.
            # In order to prevent 'results' to be empty if either the RC or FC list is empty, we added the 'dummy's.
            result = ContentEvaluationResult(
                hints={hint_key: f"Hinweis {hint_key}" for hint_key in self.hint_keys},
                # kvp is short for key value pair
                format_constraints={
                    fc_kvp[0]: EvaluatedFormatConstraint(format_constraint_fulfilled=fc_kvp[1])
                    for fc_kvp in fc_rc_tuple[0]
                    if fc_kvp[0] != "fc_dummy"
                },
                requirement_constraints={rc_kvp[0]: rc_kvp[1] for rc_kvp in fc_rc_tuple[1] if rc_kvp[0] != "rc_dummy"},
                packages={},  # is always empty because looping over all packages does not make sense in this context
            )
            if any(x for x in result.requirement_constraints.values() if x == ConditionFulfilledValue.NEUTRAL):
                continue
            results.append(result)
        return results


class CategorizedKeyExtractSchema(Schema):
    """
    A schema to (de)serialize CategorizedKeyExtractSchema
    """

    hint_keys = fields.List(fields.String())
    format_constraint_keys = fields.List(fields.String())
    requirement_constraint_keys = fields.List(fields.String())
    package_keys = fields.List(fields.String())
    time_condition_keys = fields.List(fields.String())

    @post_load
    def deserialize(self, data, **kwargs) -> CategorizedKeyExtract:
        """
        Converts the barely typed data dictionary into an actual CategorizedKeyExtractSchema
        """
        return CategorizedKeyExtract(**data)



"""
This module parses a given ahb expression like "Muss [59]U([123]O[456]) Soll [53]" or "X [59]U[53]"
using the parsing library lark: https://lark-parser.readthedocs.io/en/latest/
The goal is to separate the requirement indicator (i.e. Muss, Soll, Kann, X, O, U) from the condition expression
and also several modal marks expressions if there are more than one.
"""

from functools import lru_cache

from lark import Lark, Token, Tree
from lark.exceptions import UnexpectedCharacters, UnexpectedEOF

# pylint: disable=anomalous-backslash-in-string
from ahbicht.expressions import parsing_logger
from ahbicht.utility_functions import tree_copy

GRAMMAR = """
ahb_expression: modal_mark_expression+
                | prefix_operator_expression
                | requirement_indicator
                | modal_mark_expression+ requirement_indicator
modal_mark_expression: (MODAL_MARK CONDITION_EXPRESSION) -> single_requirement_indicator_expression
prefix_operator_expression: PREFIX_OPERATOR CONDITION_EXPRESSION -> single_requirement_indicator_expression
requirement_indicator: PREFIX_OPERATOR | MODAL_MARK
PREFIX_OPERATOR: "X"i | "O"i | "U"i
MODAL_MARK: /M(uss)?|S(oll)?|K(ann)?/i
// Matches if it looks like a condition expression, but does not yet check if it is a syntactically valid one:
CONDITION_EXPRESSION: /(?!\BU\B)[\[\]\(\)U∧O∨X⊻\d\sP\.UB]+/i
"""
# Regarding the negative lookahead in the condition expression regex see examples https://regex101.com/r/6fFHD4/1
# and CTRL+F for "Mus[2]" in the unittest that fails if you remove the lookahead.
_parser = Lark(GRAMMAR, start="ahb_expression")


@tree_copy
@lru_cache(maxsize=1024)
def parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression: str) -> Tree[Token]:
    """
    Parse a given expression as it appears in the AHB with the help of the here defined grammar to a lark tree.
    The goal is to separate the requirement indicator (i.e. Muss/M Soll/S Kann/K, X, O, U) from the condition expression
    and also several expression with modal marks if there are more than one.
    Whitespaces are ignored.

    :param ahb_expression: e.g. 'Muss[45]U[52] Soll[1]'
    :return parsed_tree:
    """
    try:
        parsed_tree = _parser.parse(ahb_expression)
        parsing_logger.debug("Successfully parsed '%s' as AHB expression", ahb_expression)
    except (UnexpectedEOF, UnexpectedCharacters, TypeError) as eof:
        # the expression is not well-formed
        parsing_logger.warning("The ahb expression is not well-formed: '%s'", ahb_expression, exc_info=eof)
        raise SyntaxError(
            f"ahb expression: {ahb_expression} "
            + """Please make sure that the ahb_expression starts with a requirement indicator \
(i.e Muss/M, Soll/S, Kann/K, X, O, U) and the condition expressions consist of only \
the following characters: [ ] ( ) U ∧ O ∨ X ⊻ and digits."""
        ) from eof

    return parsed_tree



"This module contains the classes for the validation results."

from abc import ABC
from typing import Any, Dict, List, Optional, Union

import attrs
import maus.models.edifact_components
from marshmallow import Schema, fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import-untyped]

from ahbicht.models.validation_values import RequirementValidationValue

# pylint: disable=too-few-public-methods, no-member, unused-argument


@attrs.define(auto_attribs=True, kw_only=True)
class ValidationResult(ABC):
    """Result of the validation"""

    #: In which requirement state is the field and is it filled or not?
    requirement_validation: RequirementValidationValue = attrs.field(
        validator=attrs.validators.instance_of(RequirementValidationValue)
    )
    #: Collected hints
    hints: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)), default=None
    )


class ValidationResultAttributesSchema(Schema):
    """
    A schema to pass on the attributes of ValidationResult
    """

    requirement_validation = EnumField(RequirementValidationValue)
    hints = fields.String(load_default=None)


class ValidationResultSchema(ValidationResultAttributesSchema):
    """
    A schema to (de-)serialize ValidationResult
    """

    @post_load
    def deserialize(self, data, **kwargs) -> ValidationResult:
        """
        Converts the barely typed data dictionary into an actual ValidationResult
        :param data:
        :param kwargs:
        :return:
        """
        return ValidationResult(**data)

    def dump(self, obj, **kwargs) -> Union[Any, list]:
        """
        A way to dump the subclasses DataElementValidationResult and SegmentLevelValidationResult
        of ValidationResult
        :param obj:
        :param kwargs:
        :return:
        """
        if isinstance(obj, DataElementValidationResult):
            return DataElementValidationResultSchema().dump(obj)
        if isinstance(obj, SegmentLevelValidationResult):
            return SegmentLevelValidationResultSchema().dump(obj)
        raise NotImplementedError(f"Data type of {obj} is not implemented for JSON serialization")


@attrs.define(auto_attribs=True, kw_only=True)
class SegmentLevelValidationResult(ValidationResult):
    """Result of the validation of a segment or segment group"""


class SegmentLevelValidationResultSchema(ValidationResultAttributesSchema):
    """
    A schema to (de-)serialize SegmentLevelValidationResult
    """

    @post_load
    def deserialize(self, data, **kwargs) -> SegmentLevelValidationResult:
        """
        Converts the barely typed data dictionary into an actual ValidationResult
        :param data:
        :param kwargs:
        :return:
        """
        return SegmentLevelValidationResult(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class DataElementValidationResult(ValidationResult):
    """Result of the validation of a data element"""

    #: Is the format constraint fulfilled or not?
    format_validation_fulfilled: bool = attrs.field(validator=attrs.validators.instance_of(bool))
    #: possible error message regarding the format constraints
    format_error_message: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)), default=None
    )
    #: possible qualifiers for value pool data elements
    possible_values: Optional[Dict[str, str]] = attrs.field(
        default=None,
        validator=attrs.validators.optional(
            attrs.validators.deep_mapping(
                key_validator=attrs.validators.instance_of(str),
                value_validator=attrs.validators.instance_of(str),
                mapping_validator=attrs.validators.instance_of(dict),
            )
        ),
    )
    data_element_data_type: Optional[maus.models.edifact_components.DataElementDataType] = attrs.field(
        default=None,
        validator=attrs.validators.optional(
            attrs.validators.instance_of(maus.models.edifact_components.DataElementDataType)
        ),
    )
    """
    describes the type of the original data element that has been validated
    """


class DataElementValidationResultSchema(ValidationResultAttributesSchema):
    """
    A schema to (de-)serialize DataElementValidationResult
    """

    format_validation_fulfilled = fields.Bool()
    format_error_message = fields.String(load_default=None)
    possible_values = fields.Dict(fields.Str, load_default=None)
    data_element_data_type = EnumField(
        maus.models.edifact_components.DataElementDataType, required=False, load_default=None
    )

    @post_load
    def deserialize(self, data, **kwargs) -> DataElementValidationResult:
        """
        Converts the barely typed data dictionary into an actual DataElementValidationResult
        :param data:
        :param kwargs:
        :return:
        """
        return DataElementValidationResult(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class ValidationResultInContext:
    """
    Class to set validation result in context, for example with its discriminator.
    """

    discriminator: Optional[str] = attrs.field(validator=attrs.validators.optional(attrs.validators.instance_of(str)))
    validation_result: ValidationResult = attrs.field(validator=attrs.validators.instance_of(ValidationResult))


class ValidationResultInContextSchema(Schema):
    """
    A schema to serialize ValidationResultInContext
    """

    discriminator = fields.String()
    validation_result = fields.Nested(ValidationResultSchema)

    # pylint:disable=unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> ValidationResultInContext:
        """
        Serializes a ValidationResultDiscriminator
        so that the discriminator is the key and the validation_result the value.
        """
        return ValidationResultInContext(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class ListOfValidationResultInContext:
    """
    Class to set validation result in context, for example with its discriminator.
    This class is basically a wrapper around the list to have one place where the methods operating on the list do live.
    """

    validation_results: List[ValidationResultInContext] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(ValidationResultInContext),
            iterable_validator=attrs.validators.instance_of(list),
        )
    )

    def replace_edi_domain_paths_with_application_domain_paths(
        self, edi_domain_to_application_domain_mappings: Dict[str, str]
    ) -> None:
        """
        replaces the edi domain discriminators in the list of validation results with the respective
        values from the application domain if present

        :param edi_domain_to_application_domain_mappings: dictionary with edi domain paths as keys
            and application domain paths as values, for example ediseed to bo4e mapping:
            "$[Dokument][0]['Transaktionsgrund']": "$['transaktionsdaten']['transaktionsgrund']"
        :return:
        """
        for validation_result in self.validation_results:
            if validation_result.discriminator is not None and validation_result.discriminator.startswith("$"):
                edi_seed_path = validation_result.discriminator
                if edi_seed_path in edi_domain_to_application_domain_mappings:
                    bo4e_path = edi_domain_to_application_domain_mappings[edi_seed_path]
                    validation_result.discriminator = bo4e_path

    def filter_for_data_element_validation_results(self) -> None:
        """
        Filters the lost of validation results for only DataElementValidationResults
        """
        self.validation_results = list(
            filter(lambda vr: isinstance(vr.validation_result, DataElementValidationResult), self.validation_results)
        )

    @staticmethod
    def _is_boneycomb_path_result(validation_result_in_context: ValidationResultInContext) -> bool:
        return (
            validation_result_in_context.discriminator is None
            or "stammdaten" in validation_result_in_context.discriminator
            or "transaktionsdaten" in validation_result_in_context.discriminator
        )

    def filter_for_boneycomb_path_results(self) -> None:
        """
        Filters the list of validation results for those that have a boneycomb path as discriminator
        """
        self.validation_results = list(filter(self._is_boneycomb_path_result, self.validation_results))

    @staticmethod
    def _is_not_absender_or_empfaenger_path_result(validation_result_in_context: ValidationResultInContext) -> bool:
        return (
            validation_result_in_context.discriminator is not None
            and "absender" not in validation_result_in_context.discriminator
            and "empfaenger" not in validation_result_in_context.discriminator
        )

    def remove_absender_and_empfaenger_path_results(self) -> None:
        """
        Removes all paths containing Absender oder Empfaenger
        """
        self.validation_results = list(filter(self._is_not_absender_or_empfaenger_path_result, self.validation_results))


class ListOfValidationResultInContextSchema(Schema):
    """
    A schema to deserialize ListOfValidationResultInContext
    """

    validation_results = fields.List(fields.Nested(ValidationResultInContextSchema))

    # pylint:disable=unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> ListOfValidationResultInContext:
        """
        Deserializes a ValidationResultDiscriminator
        """
        return ListOfValidationResultInContext(**data)



"""
Schemata for the JSON serialization of expressions.
"""

from lark import Tree
from marshmallow import post_dump, pre_load

from ahbicht.json_serialization.tree_schema import TreeSchema


class ConciseConditionKeyTreeSchema(TreeSchema):
    """
    A tree schema that merges condition key values and their single child
    (not deserialization)
    """

    # pylint:disable=unused-argument
    @post_dump
    def serialize(self, data, **kwargs) -> dict:
        """
        Serialize as compact/concise tree
        """
        return _compress_condition_keys_only(data)

    @pre_load
    def deserialize(self, data, **kwargs) -> Tree:
        """
        This is out of scope
        """
        raise NotImplementedError("Deserializing a compact tree is not supported.")


def _compress_condition_keys_only(data: dict) -> dict:
    """
    a function that merges a condition key node with its only child (a token that has an int value)
    """
    # this has been found heuristically. There's no way to explain it, just follow the test cases.
    # there's probably a much easier way, e.g. by using a separate token schema.
    if "tree" in data and data["tree"] is not None:
        if "type" in data["tree"]:
            if data["tree"]["type"] == "single_requirement_indicator_expression":
                if data["tree"]["children"][0]["token"]["type"] == "MODAL_MARK":
                    modal_mark = data["tree"]["children"][0]["token"]["value"]
                    del data["tree"]["children"][0]
                    data["tree"]["type"] = modal_mark
                elif data["tree"]["children"][0]["token"]["type"] == "PREFIX_OPERATOR":
                    prefix_operator = data["tree"]["children"][0]["token"]["value"]
                    del data["tree"]["children"][0]
                    data["tree"]["type"] = prefix_operator
        if data["tree"]["type"] == "condition":
            return {
                "token": {"value": data["tree"]["children"][0]["token"]["value"], "type": "condition_key"},
                "tree": None,
            }
        if "token" in data and data["token"] is None and "children" in data["tree"]:
            data["tree"]["children"] = [_compress_condition_keys_only(child) for child in data["tree"]["children"]]
    if "type" in data and data["type"] is not None and "children" in data and data["children"] is not None:
        data["children"] = [_compress_condition_keys_only(child) for child in data["children"]]
    return data



"""
A module to evaluate datetimes and whether they are "on the edge" of a German "Stromtag" or "Gastag" respectively
"""

import re
from datetime import datetime, time, timedelta
from datetime import timezone as tz
from typing import Callable, Literal, Optional, Tuple, Union

# The problem with the stdlib zoneinfo is, that the availability of timezones via ZoneInfo(zone_key) depends on the OS
# and system on which you're running it. In some cases "Europe/Berlin" might be available, but generally it's not,
# and it's PITA to manually define timezones. So we're using pytz as a datasource for timezone information.
from pytz import timezone, utc

from ahbicht import StrEnum
from ahbicht.models.condition_nodes import EvaluatedFormatConstraint

berlin = timezone("Europe/Berlin")


class EdifactDateTimeFormat(StrEnum):
    """
    EDIFACT time strings.
    """

    CCYY = "CCYY"  # 602
    CCYYMM = "CCYYMM"  # 610
    CCYYMMDD = "CCYYMMDD"  # 102
    CCYYMMDDHHMM = "CCYYMMDDHHMM"  # 203
    CCYYMMDDHHMMSS = "CCYYMMDDHHMMSS"  # 204
    CCYYMMDDHHMMZZZ = "CCYYMMDDHHMMZZZ"  # 303
    CCYYMMDDHHMMSSZZZ = "CCYYMMDDHHMMSSZZZ"  # 304
    MM = "MM"  # 802
    ZZRB = "ZZRB"  # Z01
    MMWWMMWW = "MMWWMMWW"  # 104
    MMDD = "MMDD"  # 106


def _is_edifact_time_str(time_str: str) -> bool:
    if len(time_str) == 2:
        return time_str.isdigit()
    return time_str[:-3].isdigit()


def _is_edifact_time_str_convertible_to_datetime(time_str: str) -> tuple[bool, Optional[EdifactDateTimeFormat]]:
    """
    Checks if a EDIFACT time string can be converted to datetime. If not the EdiFactDateTimeFormat is returned.
    """
    edifact_time_format: Optional[EdifactDateTimeFormat] = None
    if len(time_str) == 2:  # 802 Monat erlaubt: 1, 3, 6, 12
        edifact_time_format = EdifactDateTimeFormat.MM
    elif len(time_str) == 4 and EDIFACT_TIME_QUANTITY_REGEX.match(time_str):  # Z01 ZZRB
        edifact_time_format = EdifactDateTimeFormat.ZZRB
    elif len(time_str) == 4 and int(time_str[:2]) < 12:  # 106 MMDD -> UTILMDS
        edifact_time_format = EdifactDateTimeFormat.MMDD
    elif len(time_str) == 4:  # 602 CCYY
        edifact_time_format = EdifactDateTimeFormat.CCYY
    elif len(time_str) == 6:  # 610 CCYYMM
        edifact_time_format = EdifactDateTimeFormat.CCYYMM
    elif len(time_str) == 8 and int(time_str[:2]) < 12:  # 104 MMWWMMWW
        edifact_time_format = EdifactDateTimeFormat.MMWWMMWW
    return edifact_time_format is None, edifact_time_format


def _get_german_local_time(date_time: datetime) -> time:
    """
    returns the current german local time for the given datetime object
    """
    german_local_datetime = date_time.astimezone(berlin)
    return german_local_datetime.time()


EDIFACT_DATETIME_REGEX = re.compile(
    r"^(?P<year>\d{4})?(?P<month>\d{2})?(?P<day>\d{2})?(?P<hour>\d{2})?"
    r"(?P<minute>\d{2})?(?P<second>\d{2})?(?P<timezone>[+-]\d{2})?$"
)
EDIFACT_TIME_QUANTITY_REGEX = re.compile(r"^(?P<quantity>\d{2})(?P<unit>[TWM])(?P<reference>[MQHJTR])$")


# the functions below are excessively unit tested; Please add a test case if you suspect their behaviour to be wrong
def parse_as_datetime(entered_input: str) -> Tuple[Optional[datetime], Optional[EvaluatedFormatConstraint]]:
    """
    Try to parse the given entered_input as datetime
    :param entered_input: a string
    :return: A tuple with the first item being a datetime if parsing was successful or the second item being
    an (erroneous) EvaluatedFormatConstraint if the parsing was not successful.
    """
    if not entered_input:
        return None, EvaluatedFormatConstraint(
            format_constraint_fulfilled=False,
            error_message="An empty or None string cannot be parsed as datetime",
        )
    try:
        if entered_input.endswith("Z"):
            entered_input = entered_input.replace("Z", "+00:00")

        # check if the input could be a strict EDIFACT datetime (i.e. something like 202201010000+00)
        if _is_edifact_time_str(entered_input):
            is_convertible, edi_time_format = _is_edifact_time_str_convertible_to_datetime(entered_input)
            if not is_convertible:
                return None, EvaluatedFormatConstraint(
                    format_constraint_fulfilled=False,
                    error_message=f"EDIFACT time string cannot be parsed as datetime {entered_input}"
                    f" as it is of format {edi_time_format}.",
                )
            edifact_datetime_match = EDIFACT_DATETIME_REGEX.match(entered_input)
            if edifact_datetime_match is not None:
                groups = {
                    key: int(value)
                    for key, value in edifact_datetime_match.groupdict().items()
                    if value is not None and key != "timezone"
                }
                offset = edifact_datetime_match.groupdict().get("timezone", None)
                result = datetime(**groups, tzinfo=None if offset is None else tz(timedelta(hours=int(offset))))
            else:
                return None, EvaluatedFormatConstraint(
                    format_constraint_fulfilled=False, error_message="Could not parse EDIFACT datetime"
                )  # should not happen
        else:
            result = datetime.fromisoformat(entered_input)
        if result.tzinfo is None:
            # If tzinfo is none the datetime is "naive" = not aware of its timezone.
            # We don't want naive datetimes because they are the root of all evil.
            return None, EvaluatedFormatConstraint(
                format_constraint_fulfilled=False,
                error_message=f"Neither offset nor timezone was given: '{entered_input}'",
            )
        return result, None
    except ValueError as value_error:
        return None, EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message=str(value_error))


def has_no_utc_offset(entered_input: str) -> EvaluatedFormatConstraint:
    """
    returns true if and only if entered_input is parsable as date time with explicit offset which has no UTC offset.
    This means the UTC offset is exactly "+00:00".
    """
    # the name of the method contains a negation because this it's commonly like this in the BDEW format constraints.
    date_time, error_result = parse_as_datetime(entered_input)
    if error_result is not None:
        return error_result
    original_time = date_time.time()  # type:ignore[union-attr]
    utc_time = date_time.astimezone(tz=utc).time()  # type:ignore[union-attr]
    if utc_time == original_time and utc_time.hour == 0 and utc_time.minute == 0 and utc_time.second == 0:
        return EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None)
    error_message = f"The provided date time '{entered_input}' has a UTC offset of {utc_time}."
    return EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message=error_message)


def is_stromtag_limit(date_time: datetime) -> bool:  # the name is not as speaking as I'd like it to be
    """
    Returns true if and only if the given date_time is the inclusive start/exclusive end of a german "Stromtag".
    The "Stromtag" is the balancing relevant day in the German electricity market. It starts and ends at midnight in
    German local time which can be either 23PM or 22PM in UTC (depending on the daylight saving time in Germany).
    """
    german_local_time = _get_german_local_time(date_time)
    return german_local_time.hour == 0 and german_local_time.minute == 0 and german_local_time.second == 0


def is_gastag_limit(date_time: datetime) -> bool:  # the name is not as speaking as I'd like it to be
    """
    Returns true if and only if the given date_time is the inclusive start/exclusive end of a german "Gastag".
    The "Gastag" is the balancing relevant day in the German gas market. It starts and ends at 6am in German local time
    which can be either 4AM or 5AM in UTC (depending on the daylight saving time in Germany).
    """
    german_local_time = _get_german_local_time(date_time)
    return german_local_time.hour == 6 and german_local_time.minute == 0 and german_local_time.second == 0


def is_xtag_limit(entered_input: str, division: Union[Literal["Strom"], Literal["Gas"]]) -> EvaluatedFormatConstraint:
    """
    Tries to parse the entered_input as datetime and checks if it is the start/end of a Strom- or Gastag
    """
    date_time, error_result = parse_as_datetime(entered_input)
    if error_result is not None:
        return error_result
    xtag_evaluator: Callable[[datetime], bool]
    if division == "Strom":
        xtag_evaluator = is_stromtag_limit
    elif division == "Gas":
        xtag_evaluator = is_gastag_limit
    else:
        raise NotImplementedError(f"The division must either be 'Strom' or 'Gas': '{division}'")
    if xtag_evaluator(date_time):  # type:ignore[arg-type]
        return EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None)
    error_message = (
        f"The given datetime '{date_time.isoformat()}' is not the limit of a {division}tag"  # type:ignore[union-attr]
    )
    return EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message=error_message)



"""
Module for taking all the condition keys of a condition expression and building their respective ConditionNodes.
If necessary it evaluates the needed attributes.
"""

import sys
from typing import Dict, List, Tuple, Union

import inject

from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluatableDataProvider
from ahbicht.content_evaluation.token_logic_provider import TokenLogicProvider
from ahbicht.expressions.condition_expression_parser import extract_categorized_keys_from_tree
from ahbicht.models.condition_nodes import Hint, RequirementConstraint, UnevaluatedFormatConstraint

# TRCTransformerArgument is a union of nodes that are already evaluated from a Requirement Constraint (RC) perspective.
# The Format Constraints (FC) might still be unevaluated. That's why the return type used in the
# RequirementConstraintTransformer is always an EvaluatedComposition.
TRCTransformerArgument = Union[RequirementConstraint, UnevaluatedFormatConstraint, Hint]  # pylint:disable=invalid-name


# pylint: disable=no-member, too-few-public-methods


class ConditionNodeBuilder:
    """
    Builds ConditionNodes for the given condition_keys by separating them into their respective types
    and evaluating the necessary attributes.
    It distinguishes between requirement constraint evaluation and format constraint evaluation.
    """

    def __init__(self, condition_keys: List[str]):
        self.token_logic_provider: TokenLogicProvider = inject.instance(TokenLogicProvider)  # type:ignore[assignment]
        self.condition_keys = condition_keys
        (
            self.requirement_constraints_condition_keys,
            self.hints_condition_keys,
            self.format_constraints_condition_keys,
        ) = self._seperate_condition_keys_into_each_type()

    def _seperate_condition_keys_into_each_type(self) -> Tuple[List[str], List[str], List[str]]:
        """
        Separates the list of all condition keys into three lists of their respective types.
        """
        categorized_keys = extract_categorized_keys_from_tree(self.condition_keys)
        # note: if you remove duplicate keys or sort the keys, you'll have failing tests
        return (
            categorized_keys.requirement_constraint_keys,
            categorized_keys.hint_keys,
            categorized_keys.format_constraint_keys,
        )

    @inject.params(evaluatable_data=EvaluatableDataProvider)  # injects what has been bound to the EvaluatableData type
    # search for binder.bind_to_provider(EvaluatableDataProvider, your_function_that_returns_evaluatable_data_goes_here)
    async def _build_hint_nodes(self, evaluatable_data: EvaluatableData) -> Dict[str, Hint]:
        """Builds Hint nodes from their condition keys by getting all hint texts from the HintsProvider."""
        hints_provider = self.token_logic_provider.get_hints_provider(
            evaluatable_data.edifact_format, evaluatable_data.edifact_format_version
        )
        return await hints_provider.get_hints(self.hints_condition_keys)

    def _build_unevaluated_format_constraint_nodes(self) -> Dict[str, UnevaluatedFormatConstraint]:
        """Build unevaluated format constraint nodes."""
        unevaluated_format_constraints: Dict[str, UnevaluatedFormatConstraint] = {}
        for condition_key in self.format_constraints_condition_keys:
            unevaluated_format_constraints[condition_key] = UnevaluatedFormatConstraint(condition_key=condition_key)
        return unevaluated_format_constraints

    @inject.params(evaluatable_data=EvaluatableDataProvider)  # injects what has been bound to the EvaluatableData type
    # search for binder.bind_to_provider(EvaluatableDataProvider, your_function_that_returns_evaluatable_data_goes_here)
    async def _build_requirement_constraint_nodes(
        self, evaluatable_data: EvaluatableData
    ) -> Dict[str, RequirementConstraint]:
        """
        Build requirement constraint nodes by evaluating the constraints
        with the help of the respective Evaluator.
        """
        rc_evaluator = self.token_logic_provider.get_rc_evaluator(
            evaluatable_data.edifact_format, evaluatable_data.edifact_format_version
        )
        evaluated_conditions_fulfilled_attribute = await rc_evaluator.evaluate_conditions(
            condition_keys=self.requirement_constraints_condition_keys, evaluatable_data=evaluatable_data
        )
        evaluated_requirement_constraints: Dict[str, RequirementConstraint] = {}
        for condition_key in self.requirement_constraints_condition_keys:
            evaluated_requirement_constraints[condition_key] = RequirementConstraint(
                condition_key=condition_key,
                conditions_fulfilled=evaluated_conditions_fulfilled_attribute[condition_key],
            )
        return evaluated_requirement_constraints

    async def requirement_content_evaluation_for_all_condition_keys(self) -> Dict[str, TRCTransformerArgument]:
        """Gets input nodes for all condition keys."""
        try:
            requirement_constraint_nodes = (
                await self._build_requirement_constraint_nodes()  # pylint:disable=no-value-for-parameter
            )
            # the missing value should be injected automatically
        except AttributeError as attribute_error:
            # the 'name' attribute of the Attribute error has been added in Python3.10
            # https://docs.python.org/3/library/exceptions.html#AttributeError
            if (sys.version_info.minor < 10 or attribute_error.name == "edifact_format") and attribute_error.args[
                0
            ].startswith("'EvaluatableDataProvider' object has no attribute"):
                # This means the injection was not set up correctly.
                # Instead of the EvaluatableDataProvider being called (which would return EvaluatableData),
                # an instance of the EvaluatableDataProvider itself was instantiated.
                # Most likely you're missing binder.bind_to_provider(EvaluatableDataProvider, callable_goes_here)
                attribute_error.args = (
                    attribute_error.args[0] + ". Are you sure you called .bind_to_provider before?",
                )
            raise  # re-raise with an eventually slightly modified error message
        hint_nodes = await self._build_hint_nodes()  # pylint:disable=no-value-for-parameter
        unevaluated_format_constraint_nodes = self._build_unevaluated_format_constraint_nodes()
        input_nodes: Dict[str, TRCTransformerArgument] = {
            **requirement_constraint_nodes,
            **hint_nodes,
            **unevaluated_format_constraint_nodes,
        }
        return input_nodes



"""
instantiates a "global" logger for all parsing related stuff
"""

import logging
from typing import Optional

parsing_logger = logging.getLogger("ahbicht.expressions")
parsing_logger.setLevel(logging.DEBUG)


class InvalidExpressionError(BaseException):
    """
    Is raised when an expression is well-formed but invalid.
    A syntactical error leads to a SyntaxError during parsing with lark whereas this exception occurs during evaluation.
    """

    def __init__(self, error_message: str, invalid_expression: Optional[str] = None):
        """
        initialize the exception
        :param invalid_expression: the expression which is invalid (if known)
        :param error_message: an explanation why it is invalid
        """
        # often we don't know the invalid expression itself because we only work with the tree that after the parsing
        self.invalid_expression = invalid_expression
        self.error_message = error_message

    def __str__(self):
        if self.invalid_expression is None:
            return f"{self.__class__}: Invalid expression because: {self.error_message}"
        return f"{self.__class__}: The expression '{self.invalid_expression}' in invalid because: {self.error_message}"



"""
A module to generate ready to use evaluators.

AHBicht, in the first place, is designed to do the content evaluation itself with actual code that checks if conditions
are fulfilled, format constraints are met and gets hints from somewhere. However, AHBicht can also be used to do a
"fake" content evaluation where the single truth values of conditions are already known.
Then the approach with implementing single methods whose outcomes are already known/hardcoded at compile time
seems like a big overhead. The code representation of "all outcomes are already known" is an instance of the
ContentEvaluationResult. Now the methods below are useful. Simply provide a content evaluation result (the data) and
the evaluators are created based on the already known outcomes. You do not have to actually touch any evaluator code.
"""

from typing import Callable, Iterable, Optional, Protocol, Tuple

import inject
from efoli import EdifactFormat, EdifactFormatVersion

from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluatableDataProvider
from ahbicht.content_evaluation.fc_evaluators import (
    ContentEvaluationResultBasedFcEvaluator,
    DictBasedFcEvaluator,
    FcEvaluator,
)
from ahbicht.content_evaluation.rc_evaluators import (
    ContentEvaluationResultBasedRcEvaluator,
    DictBasedRcEvaluator,
    RcEvaluator,
)
from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider, TokenLogicProvider
from ahbicht.expressions.hints_provider import (
    ContentEvaluationResultBasedHintsProvider,
    DictBasedHintsProvider,
    HintsProvider,
)
from ahbicht.expressions.package_expansion import (
    ContentEvaluationResultBasedPackageResolver,
    DictBasedPackageResolver,
    PackageResolver,
)
from ahbicht.models.content_evaluation_result import ContentEvaluationResult


# pylint:disable=too-few-public-methods
class _HasEdifactFormatAndFormatVersion(Protocol):
    edifact_format: EdifactFormat
    edifact_format_version: EdifactFormatVersion


def _set_edifact_format_and_version(
    logic_providers: Iterable[_HasEdifactFormatAndFormatVersion],
    edifact_format: Optional[EdifactFormat],
    edifact_format_version: Optional[EdifactFormatVersion],
) -> None:
    for logic_provider in logic_providers:
        if edifact_format is not None:
            logic_provider.edifact_format = edifact_format
        if edifact_format_version is not None:
            logic_provider.edifact_format_version = edifact_format_version


def create_hardcoded_evaluators(
    content_evaluation_result: ContentEvaluationResult,
    edifact_format: Optional[EdifactFormat] = None,
    edifact_format_version: Optional[EdifactFormatVersion] = None,
) -> Tuple[RcEvaluator, FcEvaluator, HintsProvider, PackageResolver]:
    """
    Creates evaluators based on the given content_evaluation_result

    :param content_evaluation_result:
    :return:
    """
    rc_evaluator = DictBasedRcEvaluator(content_evaluation_result.requirement_constraints)
    fc_evaluator = DictBasedFcEvaluator(content_evaluation_result.format_constraints)
    hints_provider = DictBasedHintsProvider(content_evaluation_result.hints)
    package_resolver = DictBasedPackageResolver(content_evaluation_result.packages or {})
    _set_edifact_format_and_version(
        [rc_evaluator, fc_evaluator, hints_provider, package_resolver], edifact_format, edifact_format_version
    )
    return rc_evaluator, fc_evaluator, hints_provider, package_resolver


def create_content_evaluation_result_based_evaluators(
    edifact_format: Optional[EdifactFormat] = None,
    edifact_format_version: Optional[EdifactFormatVersion] = None,
) -> Tuple[RcEvaluator, FcEvaluator, HintsProvider, PackageResolver]:
    """
    Creates evaluators that expect the content evaluation result to be present in the evaluatble data
    """
    rc_evaluator = ContentEvaluationResultBasedRcEvaluator()
    fc_evaluator = ContentEvaluationResultBasedFcEvaluator()
    hints_provider = ContentEvaluationResultBasedHintsProvider()
    package_resolver = ContentEvaluationResultBasedPackageResolver()
    _set_edifact_format_and_version(
        [rc_evaluator, fc_evaluator, hints_provider, package_resolver], edifact_format, edifact_format_version
    )
    return rc_evaluator, fc_evaluator, hints_provider, package_resolver


def create_and_inject_hardcoded_evaluators(
    content_evaluation_result: ContentEvaluationResult,
    evaluatable_data_provider: Optional[Callable[[], EvaluatableData]] = None,
    edifact_format: Optional[EdifactFormat] = None,
    edifact_format_version: Optional[EdifactFormatVersion] = None,
) -> None:
    """
    Creates evaluators from hardcoded content_evaluation result and injects them

    :param content_evaluation_result:
    :return:
    """
    evaluators = create_hardcoded_evaluators(
        content_evaluation_result, edifact_format=edifact_format, edifact_format_version=edifact_format_version
    )

    def configure(binder):
        binder.bind(TokenLogicProvider, SingletonTokenLogicProvider([*evaluators]))
        if evaluatable_data_provider is not None:
            binder.bind_to_provider(EvaluatableDataProvider, evaluatable_data_provider)

    inject.configure_once(configure)



"""
contains a high-level function that checks if a given expression is valid or not.
"""

import asyncio
from typing import Any, Awaitable, Callable, List, Optional, Tuple, Union

from lark import Token, Tree
from lark.exceptions import VisitError

from ahbicht.expressions import InvalidExpressionError
from ahbicht.expressions.ahb_expression_evaluation import evaluate_ahb_expression_tree
from ahbicht.expressions.condition_expression_parser import extract_categorized_keys_from_tree
from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions
from ahbicht.models.content_evaluation_result import ContentEvaluationResult


async def is_valid_expression(
    expression_or_tree: Union[str, Tree[Token]],
    content_evaluation_result_setter: Callable[[ContentEvaluationResult], Any],
) -> Tuple[bool, Optional[str]]:
    """
    Returns true iff the given expression is both well-formed and valid.
    An expression is valid if and only if all possible content evaluations lead to a meaningful results.
    ⚠ This only works if the provided content_evaluation_result_setter writes the EvaluatableData in such a way, that
    the injected Evaluators (FC/RC/Hints/Package) can work with it.
    This is easiest for the ContentEvaluationResultBased FC/RC/Hints/Package token logic providers.
    :param content_evaluation_result_setter: a threadsafe method that writes the given Content Evaluation Result into
    the evaluatable data
    :param expression_or_tree: "Muss [1] U [2]" (returns True)  "Muss ([61]u [584]) o[583]" (returns False)
    :return: (True,None) iff the expression is valid; (False, error message) otherwise
    """
    tree: Tree[Token]
    if isinstance(expression_or_tree, str):
        try:
            tree = await parse_expression_including_unresolved_subexpressions(expression_or_tree)
        except SyntaxError as syntax_error:
            return False, str(syntax_error)
        except VisitError as visit_error:
            return False, str(visit_error)
    elif isinstance(expression_or_tree, Tree):
        tree = expression_or_tree
    else:
        raise ValueError(f"{expression_or_tree} is neither a string nor a Tree")
    categorized_key_extract = extract_categorized_keys_from_tree(tree, sanitize=True)
    evaluation_tasks: List[Awaitable] = []
    for content_evaluation_result in categorized_key_extract.generate_possible_content_evaluation_results():
        # create (but do not await) the evaluation tasks for all possible content evaluation results
        # the idea is, that if _any_ evaluation task raises an uncatched exception this can be interpreted as:
        # "the expression is invalid"
        async def evaluate_with_cer(cer: ContentEvaluationResult):
            content_evaluation_result_setter(cer)
            try:
                await evaluate_ahb_expression_tree(tree)
            except NotImplementedError as not_implemented_error:
                # we can ignore some specific errors
                if "due to missing information" in str(not_implemented_error):
                    pass  # this happens for UNKNOWN; it's okay because the expression might still be valid
                else:
                    raise not_implemented_error  # this is, in general, an indicator for an invalid expression
            except InvalidExpressionError as single_invalid_expression_error:
                if isinstance(expression_or_tree, str):
                    single_invalid_expression_error.invalid_expression = expression_or_tree
                raise

        evaluation_tasks.append(evaluate_with_cer(content_evaluation_result))
    try:
        # await the previously collected (yet unawaited) tasks
        await asyncio.gather(*evaluation_tasks)
    except InvalidExpressionError as invalid_expression_error:
        # if any evaluation throws a InvalidExpressionError the expression is invalid
        return False, invalid_expression_error.error_message
    return True, None


__all__ = ["is_valid_expression"]



"""
This module makes it possible to parse expressions including all their subexpressions, if present.
for example ahb_expressions which contain condition_expressions or condition_expressions which contain packages.
Parsing expressions that are nested into other expressions is referred to as "resolving".
"""

import asyncio
import inspect
from typing import Awaitable, List, Optional, Union, cast

import inject
from lark import Token, Transformer, Tree
from lark.exceptions import VisitError

from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluatableDataProvider
from ahbicht.content_evaluation.token_logic_provider import TokenLogicProvider
from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree
from ahbicht.expressions.package_expansion import PackageResolver
from ahbicht.models.mapping_results import Repeatability
from ahbicht.utility_functions import parse_repeatability


async def parse_expression_including_unresolved_subexpressions(
    expression: str,
    resolve_packages: bool = False,
    resolve_time_conditions: bool = True,
    replace_time_conditions: bool = True,
    include_package_repeatabilities: bool = False,
) -> Tree[Token]:
    """
    Parses expressions and resolves its subexpressions,
    for example condition_expressions in ahb_expressions or packages in condition_expressions.
    :param expression: a syntactically valid ahb_expression or condition_expression
    :param resolve_packages: if true resolves also the packages in the condition_expressions
    :param resolve_time_conditions: if true resolves also the time conditions in the condition_expressions
    :param replace_time_conditions: if true the time conditions "UBx" are replaced with format constraints
    :param include_package_repeatabilities: if true we include the repeatabilities of the packages take a look at
    PackageExpansionTransformer for a better understanding why we included this flag.
    """
    try:
        expression_tree = parse_ahb_expression_to_single_requirement_indicator_expressions(expression)
        expression_tree = AhbExpressionResolverTransformer().transform(expression_tree)
    except SyntaxError as ahb_syntax_error:
        try:
            expression_tree = parse_condition_expression_to_tree(expression)
        except SyntaxError as condition_syntax_error:
            # pylint: disable=raise-missing-from
            raise SyntaxError(f"{ahb_syntax_error.msg} {condition_syntax_error.msg}")
    if resolve_packages:
        # the condition expression inside the ahb expression has to be resolved before trying to resolve packages
        expression_tree = await expand_packages(expression_tree, include_package_repeatabilities)
    if resolve_time_conditions:
        expression_tree = expand_time_conditions(expression_tree, replace_time_conditions)
    return expression_tree


async def expand_packages(parsed_tree: Tree, include_package_repeatabilities: bool = False) -> Tree[Token]:
    """
    Replaces all the "short" packages in parser_tree with the respective "long" condition expressions
    """
    try:
        result = PackageExpansionTransformer(include_package_repeatabilities).transform(parsed_tree)
    except VisitError as visit_err:
        raise visit_err.orig_exc
    result = await _replace_sub_coroutines_with_awaited_results(result)
    return result


def expand_time_conditions(parsed_tree: Tree, replace_time_conditions: bool = True) -> Tree[Token]:
    """
    Replaces either all the "short" time conditions in parser_tree with the respective "long" condition expressions
    or replaces all the time conditions "UBx" with format constraints (and requirements constraints for UB3)
    """
    result = TimeConditionTransformer(replace_time_conditions).transform(parsed_tree)
    return result


async def _replace_sub_coroutines_with_awaited_results(tree: Union[Tree, Awaitable[Tree]]) -> Tree[Token]:
    """
    awaits all coroutines inside the tree and replaces the coroutines with their respective awaited result.
    returns an updated tree
    """
    result: Tree
    if inspect.isawaitable(tree):
        result = await tree
    else:
        # if the tree type hint is correct this is always a tree if it's not awaitable
        result = tree
    # todo: check why lark type hints state the return value of scan_values is always Iterator[str]
    sub_results = await asyncio.gather(*result.scan_values(asyncio.iscoroutine))
    for coro, sub_result in zip(result.scan_values(asyncio.iscoroutine), sub_results):
        for sub_tree in result.iter_subtrees():
            for child_index, child in enumerate(sub_tree.children):
                if child == coro:
                    sub_tree.children[child_index] = sub_result
    return result


# pylint: disable=invalid-name
# invalid-name: That's also the reason why it seemingly violates the naming conventions.
class AhbExpressionResolverTransformer(Transformer):
    """
    Resolves the condition_expressions inside an ahb_expression.
    """

    def CONDITION_EXPRESSION(self, expression):
        """
        Replacing the expression_condition with its parsed tree.
        """
        condition_tree = parse_condition_expression_to_tree(expression.value)
        return condition_tree


# pylint: disable=invalid-name
class PackageExpansionTransformer(Transformer):
    """
    The PackageExpansionTransformer expands packages inside a tree to condition expressions by using a PackageResolver.
    :param include_package_repeatabilities: Flag to include the repeatabilities of the packages.

    At this point, we implemented this flag to include a workaround for the repeatability of packages. We will convert
    the package repeatability to a condition expression and add it to the condition expression of the package (e.g.
    [2P1..2] -> [2P]U[1..2]). Thus, we can evaluate the repeatability of the package as a condition expression by
    injecting a suitable condition expression evaluator. Caution: we losen the constraints on condition keys to be able
    to that. This is a workaround and should be replaced by a more sophisticated solution which truly deals with
    package repetitions in the future (cf.  https://github.com/Hochfrequenz/ahbicht/pull/565).
    """

    def __init__(self, include_package_repeatabilities: bool = False):
        super().__init__()
        self.token_logic_provider = cast(TokenLogicProvider, inject.instance(TokenLogicProvider))
        self.include_package_repeatabilities = include_package_repeatabilities

    def package(self, tokens: List[Token]) -> Awaitable[Tree]:
        """
        try to resolve the package using the injected PackageResolver
        """
        # The grammar guarantees that there is always exactly 1 package_key token/terminal.
        # But the repeatability token is optional, so the list repeatability_tokens might contain 0 or 1 entries
        # They all come in the same `tokens` list which we split in the following two lines.
        package_key_token = [t for t in tokens if t.type == "PACKAGE_KEY"][0]
        repeatability_tokens = [t for t in tokens if t.type == "REPEATABILITY"]
        single_repeat_token = repeatability_tokens[0] if len(repeatability_tokens) == 1 else None
        # pylint: disable=unused-variable
        # we parse the repeatability, but we don't to anything with it, yet. Cf. docstring of this class.
        repeatability: Optional[Repeatability]
        if single_repeat_token:
            repeatability = parse_repeatability(single_repeat_token.value)
        else:
            repeatability = None
        return self._package_async(package_key_token, single_repeat_token)  # pylint:disable=no-value-for-parameter

    @inject.params(evaluatable_data=EvaluatableDataProvider)  # injects what has been bound to the EvaluatableData type
    # search for binder.bind_to_provider(EvaluatableDataProvider, your_function_that_returns_evaluatable_data_goes_here)
    async def _package_async(
        self, package_key_token: Token, repeatability_token: Optional[Token], evaluatable_data: EvaluatableData
    ) -> Tree[Token]:
        resolver: PackageResolver = self.token_logic_provider.get_package_resolver(
            evaluatable_data.edifact_format, evaluatable_data.edifact_format_version
        )
        resolved_package = await resolver.get_condition_expression(package_key_token.value)
        if not resolved_package.has_been_resolved_successfully():
            raise NotImplementedError(f"The package '{package_key_token.value}' could not be resolved by {resolver}")
        # the package_expression is not None because that's the definition of "has been resolved successfully"
        tree_result = parse_condition_expression_to_tree(resolved_package.package_expression)
        if self.include_package_repeatabilities and repeatability_token is not None:
            # We add the repeatability as a condition expression to the resolved package condition expression,
            # cf. docstring of this class.
            return Tree("and_composition", [tree_result, Tree(Token("RULE", "condition"), [repeatability_token])])
        return tree_result


# pylint: disable=invalid-name
class TimeConditionTransformer(Transformer):
    """
    There are two options which are chosen by the boolean `replace_time_conditions`:
    i:
    The time conditions are resolved as provided in the "Allgemeine Festlegungen".
    ii:
    The TimeConditionEvaluator replaces "time conditions" (aka "UB1", "UB2", "UB3") with evaluatable format constraints.
    This is not what BDEW suggests us to do in the "Allgemeine Festlegungen". BDEW says that "UBx" conditions have to be
    expanded just like packages: For example "UB1" shall be expanded to "([931] ∧ [932] [490]) ⊻ ([931] ∧ [933] [491])".

    This just doesn't work out for multiple reasons:

    1. Other than _all_ the other requirement constraints RC 490, 491, 492 and 493 are self-referencing.
    While "normal" RCs act like "You have to provide this ("Foo") if the other thing ("Bar") meets a requirement",
    the 49x RCs are of kind "This ("Foo") has to meet certain requirements (e.g. the end of a German day), regardless
    of the other things ("Bar"). So the usual requirement constraint evaluation approach ("Search for Bar and derive
    from there what it means for this (Foo)") won't work and is overly complicated, too.
    Also, if the pseudo requirement constraint 490-493 is UNFULFILLED, this does _not_ mean that the data must not
    be provided which is also a different behaviour compared to usual requirement constraints.

    2. No one who understands the concept of datetime+offset and is able to parse datetime+offset nowadays cares, if you
    use ("x datetime" with an "0 offset") or ("x+z datetime" that has a "z offset") instead. It's just BDEWs home-brewed
    EDIFACT serialization rule that, for no real reason, restricts us to set the offset z to "+00:00" always.
    https://imgflip.com/i/65giq5

    AHBicht won't restrict you to use datetime offset=="+00:00". Because we don't care and you should not care.
    The scope of AHBicht is to evaluate expressions and data, not to obey pointless BDEW EDIFACT rules.
    Also this default transformer won't expand the UBx conditions into something which is overly complicated and in the
    end does really fit (for above reasons).

    That being said… the thing that actually happens is:
    UBx will be replaced with a format constraint (plus a requirement constraint in case of UB3) that is fulfilled,
    if and only if the value provided obeys the original *meaning* of UBx.
    So the data provided will be evaluated just as you'd expect them to be evaluated but without all the bureaucracy.

    Condition UB1 checks if the datetime provided is the (inclusive) start / (exclusive) end of a German "Stromtag".
    Condition UB1 will be replaced with the format constraint 932.

    Condition UB2 checks if the datetime provided is the (inclusive) start / (exclusive) end of a German "Gastag".
    Condition UB2 will be replaced with the format constraint 934.

    Condition UB3 checks if the datetime provided is the (inclusive) start / (exclusive) end of a German "Stromtag" if
    the receiver is from division electricity (RC 492) or a "Gastag" if a receiver is from division gas (RC 493).
    Condition UB3 will be replaced with two XOR connected format constraints, each of them coupled to a requirement
    constraint for the respective division.
    """

    def __init__(self, replace_time_conditions: bool = True):
        super().__init__()
        self.replace_time_conditions = replace_time_conditions

    def time_condition(self, tokens: List[Token]) -> Tree:
        """
        Replace or resolve time conditions.
        """
        if self.replace_time_conditions:
            return self._replace_time_condition(tokens)
        return self._expand_time_condition(tokens)

    def _replace_time_condition(self, tokens: List[Token]) -> Tree:
        """
        Replace and resolve time conditions.
        """
        time_condition_key = tokens[0].value
        if time_condition_key == "UB1":
            # a format constraint for "Stromtag"
            return Tree("condition", [Token("CONDITION_KEY", "932")])  # we could also use 933; it doesn't matter
        if time_condition_key == "UB2":
            # a format constraint for "Gastage"
            return Tree("condition", [Token("CONDITION_KEY", "934")])  # we could also use 935; it doesn't matter
        if time_condition_key == "UB3":
            # RC 492 = receiver is from division electricity/strom
            # RC 493 = receiver is from division gas
            return parse_condition_expression_to_tree("[932][492]X[934][493]")
        raise NotImplementedError(f"The time_condition '{time_condition_key}' is not implemented")

    def _expand_time_condition(self, tokens: List[Token]) -> Tree:
        """
        try to resolve the time conditions using the injected PackageResolver
        """
        time_condition_key = tokens[0].value
        if time_condition_key == "UB1":
            # a format constraint for "Stromtag"
            return parse_condition_expression_to_tree("([931]∧[932][490])⊻([931]∧[933][491])")
        if time_condition_key == "UB2":
            # a format constraint for "Gastage"
            return parse_condition_expression_to_tree("([931]∧[934][490])⊻([931]∧[935][491])")
        if time_condition_key == "UB3":
            # RC 492 = receiver is from division electricity/strom
            # RC 493 = receiver is from division gas
            return parse_condition_expression_to_tree(
                "([931]∧[932][492]∧[490])⊻"
                "([931]∧[933][492]∧[491])⊻"
                "([931]∧[934][493]∧[490])⊻"
                "([931]∧[935][493]∧[491])"
            )
        raise NotImplementedError(f"The time_condition '{time_condition_key}' is not implemented")



"""
Evaluators are classes that evaluate AHB conditions, meaning: Based on a condition key and a message they return either
FULFILLED (true), UNFULFILLED (false) or NEUTRAL (None). Their results are used as input for the condition validation of
the entire message.
"""

import inspect
import logging
import re
from abc import ABC
from typing import Callable, Dict, Optional

import inject
from efoli import EdifactFormat, EdifactFormatVersion

# pylint: disable=too-few-public-methods


class Evaluator(ABC):
    """
    Base of all evaluators.
    To implement a content_evaluation create or update an inheriting class that has edifact_format and
    edifact_format_version set accordingly. Then create a method named "evaluate_123" where "123" is the condition key
    of the condition it evaluates.
    """

    # define some common attributes. They will be needed to find the correct validator for each use case.
    edifact_format: EdifactFormat = NotImplementedError(  # type:ignore[assignment]
        "The inheriting class needs to define a format to which it is applicable."
    )

    edifact_format_version: EdifactFormatVersion = NotImplementedError(  # type:ignore[assignment]
        "The inheriting class needs to define a format version."
    )
    _evaluation_method_name_pattern = re.compile(r"^evaluate_(?P<condition_key>\d+)$")

    def __init__(self):
        """
        initializes a cache with all evaluation methods defined in the (child) class
        """
        self._evaluation_methods: Dict[str, Callable] = {}
        self.logger: logging.Logger = logging.getLogger(self.__module__)
        self.logger.setLevel(logging.DEBUG)
        try:
            candidates = inspect.getmembers(self, inspect.ismethod)
        except inject.InjectorException as injector_exception:
            if str(injector_exception) == "No injector is configured":
                # this is due to https://github.com/ivankorobkov/python-inject/issues/77
                # remove this catch block once the original issue is resolved
                raise AttributeError(
                    f"Using inject.attr in custom evaluators as you tried in {self.__class__.__name__} is not "
                    "supported. Try inject.instance in any method except for __init__ instead."
                ) from injector_exception
            raise
        for candidate in candidates:
            # a candidate is a tuple of a string (index 0, name of the method) and the bound method itself (index 1)
            match = Evaluator._evaluation_method_name_pattern.match(candidate[0])
            if match:
                self._evaluation_methods[match.groupdict()["condition_key"]] = candidate[1]
        self.logger.info(
            "Instantiated %s and found %i evaluation methods", self.__class__.__name__, len(self._evaluation_methods)
        )

    def get_evaluation_method(self, condition_key: str) -> Optional[Callable]:
        """
        Returns the method that evaluates the condition with key condition_key
        :param condition_key: unique key of the condition, e.g. "59"
        :return: The method that can be used for content_evaluation; None if no such method is implemented.
        """
        if condition_key in self._evaluation_methods:
            return self._evaluation_methods[condition_key]
        return None



"""
A module to allow easy distinction between different types of condition nodes (by mapping their integer key)
"""

import re

from ahbicht.models.condition_node_type import ConditionNodeType

REGEX_PACKAGE_REPEATABILITY = re.compile(r"^(?P<n>\d)\.\.(?P<m>\d)$")


def derive_condition_node_type(condition_key: str) -> ConditionNodeType:
    """
    Returns the corresponding type of condition node for a given condition key
    """
    if condition_key.endswith("P"):
        return ConditionNodeType.PACKAGE
    match = REGEX_PACKAGE_REPEATABILITY.match(condition_key)
    if match and int(match.group("n")) <= int(match.group("m")):
        return ConditionNodeType.PACKAGE_REPEATABILITY
    if 1 <= int(condition_key) <= 499:
        return ConditionNodeType.REQUIREMENT_CONSTRAINT
    if 500 <= int(condition_key) <= 900:
        return ConditionNodeType.HINT
    if 901 <= int(condition_key) <= 999:
        return ConditionNodeType.FORMAT_CONSTRAINT
    if 2000 <= int(condition_key) <= 2499:
        return ConditionNodeType.REPEATABILITY_CONSTRAINT
    raise ValueError("Condition key is not in valid number range.")
